#!/usr/bin/perl -sw
##
## Razor2::Client::Agent -- UI routines for razor agents.
##
## Copyright (c) 2002, Vipul Ved Prakash.  All rights reserved.
## This code is free software; you can redistribute it and/or modify
## it under the same terms as Perl itself.
##
## $Id: Agent.pm,v 1.29 2002/06/16 00:07:13 chad Exp $

package Razor2::Client::Agent;

use lib qw(lib);
use strict;
use Getopt::Long; 

use base qw(Razor2::Client::Core);
use base qw(Razor2::Client::Config);
use base qw(Razor2::Logger);
use base qw(Razor2::String); 
use Razor2::Preproc::Manager;
use Razor2::Preproc::VR1;
use vars qw( $VERSION $PROTOCOL );



$PROTOCOL = $Razor2::Client::Version::PROTOCOL;
$VERSION  = $Razor2::Client::Version::VERSION;




sub new {
    my ($class, $breed) = @_;

    my @valid_program_names = qw(
            razor-check 
            razor-report  
            razor-revoke  
            razor-register
    );

    my $ok = 0;
    foreach (@valid_program_names) { $breed =~ /$_$/ and $ok = $_; }
    unless ($ok) {
        if ($breed =~ /razor-client$/) {
            create_symlinks(@valid_program_names);
            exit 0;
        }
        die "Invalid program name, must be one of: @valid_program_names\n";
    }

    $ok =~ /razor-(.*)$/; 
    my %me = (
            name_version => "Razor-Agents v$VERSION",  # used in register
            breed        => $1,
            preproc_vr1  => new Razor2::Preproc::VR1,
            preproc      => new Razor2::Preproc::Manager,
    );

    return bless \%me, $class;
}

sub do_conf {
    my $self = shift;

    return unless $self->read_conf();

    my $logto = $self->{opt}->{debug} ? "stdout" : "file:$self->{conf}->{logfile}";
    if (exists $self->{conf}->{logfile}) {
        my $debuglevel = exists $self->{conf}->{debuglevel} ? $self->{conf}->{debuglevel} : 18;
        my $logger = new Razor2::Logger ( 
                        LogDebugLevel => $debuglevel,
                        LogTo         => $logto,
                        LogPrefix     => $self->{breed},
                        LogTimestamp  => 1,
                     );
        $self->{logref} = $logger;
    }
    return 1;
}

# wrapper for log
sub log {
    my $self = shift;
    return unless $self->{logref};
    return        $self->{logref}->log(@_);
}
sub log2file {
    my $self = shift;
    return unless $self->{logref};
    return        $self->{logref}->log2file(@_);
}

sub doit {
    my $self = shift;
    my $r;

    $self->log(5,"$self->{name_version} starting razor-$self->{breed}");
    $r = $self->checkit    if $self->{breed} eq 'check';
    $r = $self->registerit if $self->{breed} eq 'register';
    $r = $self->reportit   if $self->{breed} eq 'report';
    $r = $self->reportit   if $self->{breed} eq 'revoke';

    return $r if $r;

    my $msg = $self->errstr;
    $self->log(1,"razor-$self->{breed} error: ". $msg);
    return;
}


sub _help {
    my ($self,$breed) = @_;

    chomp(my $all = <<EOFALL);
            -h  Print this usage message.
            -v  Print version number and exit
            -d  Turn on debugging.  Logs to stdout.
            -s  Simulate Only.  Does not connect to server.
    -conf file  Use this config file instead of <razorhome>/razor.conf
           -rs  Use this razor server instead of reading .lst
EOFALL
    chomp(my $sigs = <<EOFSIGS);
            -H  Compute and print signature.
   -S |  --sig  Accept a signatures to check on the command line
        -e eng  Engine used to compute sig, integer
      -ep4 val  String value required when engine == 4
EOFSIGS

    chomp(my $mbox = <<EOFMBOX);
   -M | --mbox  Accept a mailbox name on the command line (default)
                If no filename, mbox, or signatures, input read from stdin.
EOFMBOX
    
    my %b;
    $b{check} = <<EOFCHECK;

razor-check [options] [ filename | -M mbox | -S signatures | < filename ]
$all
$sigs
$mbox

See razor-check(1) manpage for details.

EOFCHECK

    $b{report} = <<EOFREPORT;

razor-report [options] [ filename | -M mbox | -S signatures -e engine]
$all
$sigs
$mbox
       -i file  Use identity from this file
            -f  Stay in foreground.
            -a  Authenticate only.  Exit 0 if authenticated, 1 if not
                Stays in foreground.

See razor-report(1) manpage for details.

EOFREPORT

    $b{register} = <<EOFREGISTER;

razor-register [options] 
$all
    -user name  Request 'name' when registering
    -pass pass  Request 'password' when registering
       -create  Initialize razorhome (~/.razor), copying from /etc/razor if it exists.
                Does not connect to server. (implies -s)
          -sys  Create system-wide conf (in /etc/razor).
                Does not connect to server. (implies -s)
            -l  Make new identity the the default identity.
                Used only when registering.

See razor-register(1) manpage for details.

EOFREGISTER

    $b{revoke} = <<EOFREVOKE;

razor-revoke [options] filename
$all
$mbox
       -i file  Use identity from this file
            -f  Stay in foreground.
            -a  Authenticate only.  exit 0 if authenticated, 1 if not
                Stays in foreground.

See razor-revoke(1) manpage for details.

EOFREVOKE

    my $future = <<EOFFUTURE;
EOFFUTURE

    return $b{$self->{breed}};
}


# maybe this should be in Client::Config
#
sub read_options { 
    my ($self, $agent) = @_;
    Getopt::Long::Configure ("no_ignore_case");
    my %opt;
    #
    # These options override what is loaded in config file
    # the names on the right should match keys in config file
    #
    GetOptions(
        's'   => \$opt{simulate},
        'd'   => \$opt{debug},
  'verbose'   => \$opt{debug},
        'v'   => \$opt{version},
        'h'   => \$opt{usage},
     'help'   => \$opt{usage},
        'H'   => \$opt{printhash},
      'C=s'   => \$opt{printcleaned},
    'sig=s'   => \$opt{sig},
      'S=s'   => \$opt{sig},
      'e=s'   => \$opt{sigengine},
    'ep4=s'   => \$opt{sigep4},
     'mbox'   => \$opt{mbox},
        'M'   => \$opt{mbox},
        'n'   => \$opt{negative},
   'conf=s'   => \$opt{config},
 'config=s'   => \$opt{config},
        'f'   => \$opt{foreground},
     'noml'   => \$opt{noml},
   'user=s'   => \$opt{user},
      'u=s'   => \$opt{user},
   'pass=s'   => \$opt{pass},
        'a'   => \$opt{authen_only},
     'rs=s'   => \$opt{server},
 'server=s'   => \$opt{server},
        'l'   => \$opt{symlink},
      'i=s'   => \$opt{identity},
      'sys'   => \$opt{sys_conf},
   'create'   => \$opt{create_conf},
    );

    if ($opt{create_conf} || $opt{sys_conf}) { 
        $opt{simulate} = 2;
    }
    if ($opt{usage}) { 
        $self->error($self->_help);
        return;
    } elsif ($opt{mbox} && $opt{sig}) { 
        $self->error("--mbox and --sig are mutually exclusive.\n"); 
        return;
    } elsif ($opt{create_conf} && $opt{user}) { 
        $self->error("-user is for register, -create is not, they are mutually exclusive.\n"); 
        return;
        # fixme - add other mut. exclisive cases
    } elsif ($opt{sig} && !$opt{sigengine}) { 
        $self->error("--sig requires -e (engine used to generate sig)\n");
        return;
        #
        # fixme - require ep4 if -e 4 is used ?
        #
    } elsif ($opt{version}) { 
        $self->error("Razor 2 Client Tools $Razor2::Client::Version::VERSION, protocol version $Razor2::Client::Version::PROTOCOL\n");
        return;
    } 
    $self->{opt} = \%opt;
    return 1;
} 




sub checkit {
    my $self = shift;

    # check for spam.
    # input can be one of 
    #   file - single mail
    #   mbox - many  mail
    #   sig  - 1 or more sigs
        
    my $objects;
    if ($self->{conf}->{sig}) {
        my @sigs;
        # 
        # cmd-line sigs
        # 
        # prepare 1 mail object per sig 
        # 
        foreach my $sig (split ',', $self->{conf}->{sig}) {
            $sig =~ s/^\s*//;  $sig =~ s/\s*$//;
            my $hr = { 
                eng => $self->{conf}->{sigengine},
                sigs => [$sig],
            };
            $hr->{ep4} = $self->{conf}->{sigep4} if $self->{conf}->{sigep4};
            push @sigs, $hr;
        }
        $self->log (5,"received ". (scalar @sigs) ." valid cmd-line sigs.");
        $objects = $self->prepare_objects(\@sigs) or return;
    } else {

        my $mails = $self->parse_mbox() or return;
        $objects  = $self->prepare_objects($mails) or return;
        #
        # if mail is whitelisted, its not spam.
        # flag it so it we don't check it against server
        #
        foreach my $obj (@$objects) {
            if ($self->local_check($obj)) {
                $obj->{skipme} = 1;
                $obj->{spam} = 0;
            } else {
                next;
            }
        }

    }

    # compute_sigs needs server info like ep4, so get_server_info first
    $self->get_server_info()     or return $self->errprefix("checkit");
    my $printable_sigs = $self->compute_sigs($objects) or return $self->errprefix("checkit");

    if ($self->{opt}->{printhash}) { 
        my $i = 0;
        foreach (@$printable_sigs) {
            if ($self->{opt}->{sigengine}) {
                next unless / e$self->{opt}->{sigengine}: /;
            }
            print "$_\n";
            $i++;
        }
        $self->log (4, "Done. Printed $i sig(s) for ". scalar(@$objects) ." mail(s)");
    }
    if ($self->{opt}->{printcleaned}) { 
        my $totalp = 0;
        my $totalc = 0;
        foreach my $obj (@$objects) {
            my $n = 0;
            `mkdir -p $self->{opt}->{printcleaned}/cleaned`;
            foreach ($obj->{headers}, @{$obj->{bodyparts_cleaned}}) {
                my $fn = "$self->{opt}->{printcleaned}/cleaned/mail$obj->{id}.". $n++;
                open OUT, ">$fn" or die "can't write to $fn";
                print OUT $$_;
                close OUT;
                $totalc++;
            }
            $n = 0;
            `mkdir -p $self->{opt}->{printcleaned}/uncleaned`;
            foreach ($obj->{headers}, @{$obj->{bodyparts}}) {
                my $fn = "$self->{opt}->{printcleaned}/uncleaned/mail$obj->{id}.". $n++;
                open OUT, ">$fn" or die "can't write to $fn";
                print OUT $$_;
                close OUT;
                $totalp++;
            }
        }
        $self->log (4, "Done. $totalp uncleaned, $totalc cleaned mails saved in $self->{opt}->{printcleaned}");
        print "Done. $totalp uncleaned, $totalc cleaned mails saved in $self->{opt}->{printcleaned}\n";
        return 1;
            
    }
    return 1 if $self->{opt}->{printhash};



    # only check good objects
    my @goodones;
    foreach my $obj (@$objects) {
        next if $obj->{skipme};
        push @goodones, $obj;
    }
    unless (scalar @goodones) {
        $self->log (4,"Done.  No valid mail or signatures to check.");
        exit 1;
    }

    if ($self->{conf}->{simulate}) {
        $self->log (4, "Done. (simulate only)");
        return 1;
    }
    #
    # Check against server
    #  
    $self->connect()          or return $self->errprefix("checkit");
    $self->check (\@goodones) or return $self->errprefix("checkit");
    $self->disconnect()       or return $self->errprefix("checkit");


    #
    # print out responses and exit
    #  
    my $only1check = (scalar(@$objects) == 1) ? 1 : 0;
    my $has_spam = 0;
    foreach my $obj (@$objects) {

        if ($obj->{spam} == 1) {
            exit 0 if $only1check;
            $has_spam = 1;
            print $obj->{id} ."\n";
            next;

        } elsif ($obj->{spam} == 0) {
            exit 1 if $only1check;
            print "-". $obj->{id} ."\n" if $self->{conf}->{negative};
            next;

        } else {
            # error
            #
            exit 1 if $only1check;
            print "-". $obj->{id} ."\n" if $self->{conf}->{negative};
            $self->logobj(1,"bad 'spam' in checkit", $obj);
            next;
        }
    }
    exit 0 if $has_spam;
    exit 1;
}




sub registerit { 
    my $self = shift;

    my $ident;
    if ($self->{opt}->{create_conf}) {
        $self->write_conf();
    }

    if (exists $self->{opt}->{user} 
        && ($ident = $self->get_ident) 
        && $ident->{user} eq $self->{opt}->{user} ) {
        return $self->error("You are already registered as user=$ident->{user} in $self->{razorhome}");
    }
    if ( $self->{conf}->{simulate}) {
        $self->log(5,"Done - simulate only.");
        exit 0;
    }

    $self->get_server_info()    or return $self->errprefix("registerit");
    $self->connect()            or return $self->errprefix("registerit");
    $ident = $self->register({
                        user => $self->{opt}->{user}, 
                        pass => $self->{opt}->{pass}, 
                            }) or return $self->errprefix("registerit");
    $self->disconnect()         or return $self->errprefix("registerit");

    if (my $fn = $self->save_ident($ident)) {
        print "Register successful.  Identity stored in $fn\n";
        exit 0;
    } else {
        print "Register unsuccessful: ". $self->errstr ."\n";
        exit 1;
    }
}


#
# handles report and revoke
#
sub reportit {
    my $self = shift;

    my $ident = $self->get_ident or return;

    if ($self->{opt}->{authen_only}) { 
        $self->authenticate($ident) or return;
        $self->log(5,"Done - authenticate only.");
        exit 0 if $self->{authenticated};
        exit 1;
    }

    my $mails   = $self->parse_mbox() or return;

    # background myself 
    unless ($self->{opt}->{foreground}) {
        chdir '/'; fork && exit; POSIX::setsid;
        # close 0, 1, 2;
    }

    my $objects = $self->prepare_objects($mails) or return;  

    # compute_sigs needs server info like ep4, so get_server_info first
    $self->get_server_info()    or return $self->errprefix("reportit");
    my $printable_sigs = $self->compute_sigs($objects) or return $self->errprefix("reportit");

    if ($self->{opt}->{printhash}) { 
        foreach (@$printable_sigs) {
            if ($self->{opt}->{sigengine}) {
                next unless / e$self->{opt}->{sigengine}: /;
            }
            print "$_\n";
        }
        exit 0;
    }

    if ( $self->{conf}->{simulate}) {
        $self->log (4, "Done. (simulate only)");
        exit 0;
    }
    unless (scalar @$objects) {
        $self->log (4,"Done.  No valid mail or signatures to check.");
        exit 1;
    }

    $self->connect()            or return $self->errprefix("reportit");
    $self->authenticate($ident) or return $self->errprefix("reportit");
    $self->report($objects)     or return $self->errprefix("reportit");
    $self->disconnect()         or return $self->errprefix("reportit");


    if ($self->{opt}->{foreground}) {
        foreach my $obj (@$objects) {
            # my $line = debugobj($obj->{r});
            # $line =~ /(\S+=\S+)/s;  # could be res=0|1, err=xxx
            # print "$obj->{id}: $1\n";
            print "$obj->{id}\n" if $obj->{r}->{res} == '1';
        }
    } else {
        return 1;
    }
}


sub parse_mbox {
    my ($self) = @_;

    my @mails;
    my @message;
    
    # There are different kinds of mbox formats, we just split on simplest case.
    # djb defines mbox, mboxrd, mboxcl, mboxcl2 
    # http://www.qmail.org/qmail-manual-html/man5/mbox.html
    #
    while (1) {
        my $line = <>; # reads from stdin or file - $ARGV[0]
        if (($line =~ /^From / && scalar @message) || !$line) {
            # next if From doesn't look like a proper from
            #next unless ($line && $line =~ /^From (\S+)\s+(.{25})/);
            my $one_mail = join '', @message;
            #print "length=". length($one_mail) ."\n";  # mutt chomp bug
            push @mails, \$one_mail if $one_mail;
            splice @message;
            last unless $line;
        }
        push @message, $line if $line;
    }

    my $cnt = scalar @mails; 
    $self->log (6, "read $cnt mail". ($cnt>1 ? 's' : '') );

    return \@mails;
}



sub raise_error { 
    my ($self, $errstr) = @_;;
    my $str;
    if (ref $self) { 
        $str = $self->errstr;
    } 
    $str = $errstr if $errstr;
    my ($code) = $str =~ /Razor Error (\d+):/;
    $code = 255 unless $code; 
    print "FATAL: $str";
    exit $code;
}

# returns 1 if mail should be skipped 
#
sub local_check {
    my ($self, $obj) = @_;
    my ($headers, $body) = split /\n\r*\n/, ${$obj->{orig_mail}}, 2;

    if ($self->{conf}->{ignorelist}) { 
        if ($headers =~ /\n((X-)?List-Id[^\n]+)/i) { 
            my $listid = $1;
            my ($line1) = substr(${$obj->{orig_mail}}, 0, 50) =~ /^([^\n]+)/;
            $self->log (5,"Mailing List post; mail ". $obj->{id} ." not spam.");
           #$self->log (5,"Mailing List post; mail ". $obj->{id} ." not spam.\n  $line1\n  $listid");
            return 1;
        }
    }
    if (-s $self->{conf}->{whitelist}) {  
        my $whitelist = $self->read_whitelist;
        my $iswl = 0;
        foreach my $sh (keys %$whitelist) { 
            if ($sh ne 'sha1') { 
                my $fc = $headers; 
                next unless $fc;
                my @list = @{$$whitelist{$sh}};
                for (@list) {
                    my $address = $_; 
                    if ($fc =~ /$address/i) { 
                        $self->log (4,"Address <$address> is whitelisted; ignoring mail ". $obj->{id});
                        return 1;
                    }
                }
            }
        }
    }
    return 0;
}



sub read_whitelist { 
    my ($self) = @_; 
    my %whitelist;
    my $whitelist = $self->{conf}->{whitelist}; 
    unless (open WL, $whitelist) { 
        $self->debug ("Unable to open the $self->{whitelist}\n");
        exit 1; 
    }
    my @WL = <WL>;
    close WL;
    for (@WL) { 
        s/^\s*//;
        next if /^#/;
        chomp;
        my ($type, $value) = split /\s+/, $_, 2; 
        $type =~ y/A-Z/a-z/ if $type;
        push @{$whitelist{$type}}, $value if ($type && $value);
    }
    return \%whitelist;
}   


sub logerr {
    my ($self,$msg) = @_; 
    $msg = $self->errstr unless $msg;
    $self->log(1,"$self->{breed} error: ". $msg);
    return;
}



sub create_symlinks {
    my @progs = @_;
    my $bdir = $0 =~ /^\// ? $0 : "$ENV{PWD}/$0";
    $bdir =~ s|/([^/]+)$||;
    foreach (@progs) {
        if (-e "$bdir/$_") {
            next if -l "$bdir/$_";           # unless a symlink, prolly razor2
            rename ("$bdir/$_", "$bdir/$_.v1")  # rename razor1 
        }
        print "Creating symlink $1 <== $bdir/$_\n";
        symlink $1, "$bdir/$_";
    }
}


sub get_server_info {
    my $self = shift;

    if ($self->{conf}->{server}) {  # cmd-line
        $self->{s}->{list} = [$self->{conf}->{server}];
    } else {
        $self->readservers;
    }
    $self->loadservercache;
    #$self->logobj(6,"find_closest_server server info (before nextserver)", $self->{s});
    $self->{loaded_servers} = 1;
    return $self->nextserver;
}


# see nextserver() for explanation of how data is stored
#
sub readservers {
    my $self = shift;

    unless (exists $self->{s}) { $self->{s} = {}; }
    unless (exists $self->{s}->{listfile}) { 
        $self->{s}->{listfile} = $self->{breed} =~ /^check/ ?
            $self->{conf}->{listfile_catalogue} : $self->{conf}->{listfile_nomination};
    }
    $self->{s}->{dsfile} = $self->{conf}->{listfile_discovery} unless $self->{s}->{dsfile};

    # read 'listfile' (nomination or catalogue)
    # and 'dsfile' 
    foreach my $fn ("list", "ds") {
        my $h = $self->read_file($self->{s}->{$fn."file"},0,1) or next;
        $self->{s}->{$fn} = [] unless $self->{s}->{$fn};
        foreach (keys %$h) {
            $self->{s}->{$fn} = [@{$self->{s}->{$fn}}, $1]
                if /^(([^\.\s]+\.)+[^\.\s]+(:\S+)?)/;  
        }
        $self->log(6,"Read ". scalar(@{$self->{s}->{$fn}}) ." from server listfile: ".
            $self->{s}->{$fn."file"});
    }
}

sub loadservercache {
    my $self = shift;

    #
    # Read in server-specific config, using defaults for stuff not found
    #
    # NOTE: this reads all server.*.conf files in razor home, not just those in .lst
    #
    foreach (@{$self->{s}->{list}}) {
        $self->{s}->{allconfs}->{$_} = $self->default_server_conf();
    }
    my @fns;
    if (opendir D,$self->{razorhome}) {
        @fns = map "$self->{razorhome}/$_", grep /^server\.[\S]+\.conf$/, readdir D;
        closedir D;
    }
    foreach (@fns) {
        /server\.(.+)\.conf$/ and my $sn = $1;
        $self->{s}->{allconfs}->{$sn} = $self->read_file($_, $self->{s}->{allconfs}->{$sn} );
        if ($self->{s}->{allconfs}->{$sn}) {
            $self->log(8,"Loaded server specific conf info for $sn");
        } else {
            $self->log(5,"loadservercache skipping $_");
        }
    }

    return $self;
}


sub writeservers {
    my $self = shift;

    if ($self->{s}->{listfile} && $self->{s}->{list} && $self->{done_discovery}) {
        if (open LIST, ">$self->{s}->{listfile}") {
            my $total = 0;
            foreach (@{$self->{s}->{list}}) {
                next unless /\S/; 
                print LIST "$_\n";
                $total++;
            }
            close LIST;
            $self->debug("Wrote $total to server list file: $self->{s}->{listfile}");
        } else {
            $self->log(4,"Couldn't open $self->{s}->{listfile} for writing: $!");
        }
    } else {
        $self->debug("Error writeservers: - Missing server info") 
            if $self->{done_discovery};
    }
    if ($self->{s}->{dsfile} && $self->{s}->{ds} && $self->{s}->{done_bootstrap}) {
        if (open LIST, ">$self->{s}->{dsfile}") {
            my $total = 0;
            foreach (@{$self->{s}->{ds}}) {
                next unless /\S/; 
                print LIST "$_\n";
                $total++;
            }
            close LIST;
            $self->debug("Wrote $total to discovery server list file: $self->{s}->{dsfile}");
        } else {
            $self->log(4,"Couldn't open $self->{s}->{dsfile} for writing: $!");
        }
    }
    foreach (@{$self->{s}->{modified}}) {
        my $fn = "$self->{razorhome}/server.$_.conf";
        my $header = "#\n# Autogenerated by $self->{name_version}, ". localtime() ."\n";
        $header .= "# Any changes will be lost.\n#\n";
        $self->write_file($fn, $self->{s}->{allconfs}->{$_}, 0, $header) 
            || $self->debug("writeservers skipping $fn");
    }
    return $self;
}



1;
