#!/usr/bin/perl -s
## 
## Razor2::Client:Config
## 
## Copyright (c) 2002, Vipul Ved Prakash.  All rights reserved.
## This code is free software; you can redistribute it and/or modify
## it under the same terms as Perl itself.
##
## $Id: Config.pm,v 1.21 2002/06/16 00:32:31 vipul Exp $

package Razor2::Client::Config; 
use strict;
use Data::Dumper;
use vars qw( $VERSION ); 
use File::Copy;

use Razor2::Logger;

# where global razor-agent.conf is
my $sys_conf_dir = "/etc/razor";

#use base qw(Razor2::Logger);

sub new { 
    my ($class) = @_; 
    return bless {}, $class;
} 


sub read_conf {
    my ($self,$params) = @_;

    my $conf;
    my $defaults = $self->default_agent_conf();

    if ($self->{opt}->{config}) {
        die "Can't read config file $self->{opt}->{config}: $!\n" unless -r $self->{opt}->{config};
        $self->{razorconf} = $self->{opt}->{config};
        $conf = $self->read_file($self->{razorconf},$defaults);
        $self->find_home($conf->{razorhome}) or return;
    } else {
        $self->find_home() or return;

        # figure out which conf to use - user's own, or system conf dir (/etc/razor)
        #
        # NOTE: if user is root and invoked sys_conf, razorhome is sys_conf_dir
        my $mycf = "$self->{razorhome}/razor-agent.conf";
        my $syscf = "$sys_conf_dir/razor-agent.conf";
        my $foundconf = "";
        $self->{razorconf} = $mycf;
        if (-e $mycf) {
            if (-r $mycf) {
                $foundconf = $mycf;
            } else {
                $self->log(5, "Found but can't read $mycf, skipping.");
            }
        }
        if (!$foundconf && -e $syscf) {
            if (-r $syscf) {
                $foundconf = $syscf;
            } else {
                $self->log(5, "Found but can't read $syscf, skipping.");
            }
        }
        if ($foundconf) {
            $conf = $self->read_file($foundconf,$defaults);
        } else {
            $self->log(6, "No razor-agent.conf found, using defaults.  Please razor-register -create");
            $conf = $defaults;
        }
    }

    foreach (keys %{$self->{opt}}) {
        next if ($_ eq '');
        $conf->{$_} = $self->{opt}->{$_};
    }
    if ($params) {
        foreach (keys %$params) {
            $conf->{$_} = $params->{$_};
        }
    }
    $self->{conf} = $conf; 

    
    #
    # post config processing
    # insert things that should not be in conf here
    #

    # add full path to all config values that need them
    #
    foreach (qw( logfile pidfile listfile_catalogue listfile_nomination
                 listfile_discovery whitelist )) {
        next unless $conf->{$_};
        next if $conf->{$_} =~ /^\//;
        $conf->{$_} = "$self->{razorhome}/$conf->{$_}";
    }

    return $self->{conf}; 
} 



sub write_conf {
    my ($self,$hash)  = @_;
    my $now = localtime();
    my $srcmsg;
    unless ($hash) {
        $hash = $self->default_agent_conf();
        if (-r "$sys_conf_dir/razor-agent.conf") {
            $hash = $self->read_file( "$sys_conf_dir/razor-agent.conf", $hash);
            $srcmsg = "Non-default values taken from $sys_conf_dir/razor-agent.conf";
        } else {
            $hash = $self->default_agent_conf();
            $srcmsg = "Created with all default values";
        }
    }

    my $clientheader = <<EOFCLIENT; 
#
# Razor2 config file
# 
# Autogenerated by $self->{name_version} 
# $now
# $srcmsg 
# 
# see razor-agent.conf(5) man page 
#
EOFCLIENT
    return $self->write_file($self->{razorconf}, $hash, 0, $clientheader);
}

sub find_user {
    my $self = shift;

    return 1 if $self->{user};
    $self->{user} = getlogin() || getpwuid($>) || do {
        $self->log(1, "Can't figure out who the effective user is: $!");
        return undef;
    };
    return 1;
}

sub find_home {
    my ($self,$rhome) = @_;

    return 1 if defined $self->{razorhome};

    # user <home>/.razor unless sys_conf && root, then $sys_conf_dir
    #
    unless ($rhome) {
        if (defined $ENV{HOME}) {
            $rhome = "$ENV{HOME}/.razor";
        } else {
            return undef unless $self->find_user();
            $rhome = (getpwnam($self->{user}))[7] . "/.razor" || "/home/$self->{user}/.razor";
        }
        if ($self->{opt}->{sys_conf}) {
            $self->find_user();
            if (($self->{user} eq 'root') || (getpwuid($>) eq 'root')) {
                $rhome = $sys_conf_dir;
            } else {
                $self->log(5,"non-root cannot install to $sys_conf_dir, using $rhome");
            }
        }

    }
    $self->{razorhome} = $rhome;

    return 1 if -d $rhome; # fixme - check to see if i can read/write to dir?

    return 1 if mkdir $rhome, 0755;

    $self->error("Can't mkdir $rhome: $!");
    return;
}


sub get_ident {
    my ($self) = @_;
    $self->find_home() or return;

    my $idfn = $self->{opt}->{identity} || "$self->{razorhome}/identity";
    $idfn = readlink($idfn) if -l $idfn;
    $idfn = "$self->{razorhome}/$idfn" unless $idfn =~ /^\//;

    my $mode = ((stat($idfn))[2]) & 07777; # mask off file type
    if ($mode & 0007) {
        return $self->error("Invalid file permissions.  Please chmod 700 $idfn");
    }
    return $self->read_file( $idfn );
}

sub save_ident {
    my ($self,$ident) = @_;
    $self->find_home() or return;

    my $obase = "identity-$ident->{user}";
    my $orig  = "$self->{razorhome}/$obase";
    my $syml  = "$self->{razorhome}/identity";  

    my $umask = umask 0077;  # disable group and all from read/write/execute 
    $self->write_file($orig,$ident) or return;
    umask $umask;

    unless ($self->{opt}->{symlink}) {
        return $orig if -e $syml;  # already has another identity
    }

    unlink $syml if -e $syml;
    if (eval { symlink("",""); 1 } ) {
        symlink $obase, $syml or
            return $self->error("Created $orig, but could not symlink to it $syml: $!");
    } else {
        $self->log(5, "symlinks dont work on this machine");
        copy($orig,$syml);
    }
    return $orig;
}

sub parse_value {
    my ($self, $value) = @_;

    $value =~ s/^\s+//; 
    $value =~ s/\s+$//;
    if ($value =~ m:,:) { 
        my @values = split /,\s*/, $value;
        return [@values];
    } else { 
        return $value;
    }
}

# given filename, returns hash ref of key = val from file 
# if $serverfile, than no key && val, just keys.
#
sub read_file {
    my ($self,$fn,$h,$serverfile) = @_;

    my $conf = ref($h) eq 'HASH' ? $h : {};

    my $total = 0;
    open(CONF, "<$fn") or 
        return $self->error("Can't read file $fn: $!");
    for (<CONF>) { 
        chomp; 
        next if /^\s*#/;
        if ($serverfile) {
            s/^\s+//; s/\s+$//;
            $conf->{$_} = 7;
        } else { 
            next unless /=/;
            my ($attribute, $value) = split /\=/, $_, 2; 
            $attribute =~ s/^\s+//; $attribute =~ s/\s+$//;
            $conf->{$attribute} = $self->parse_value($value);
        }
        $total++;
    }
    close CONF;
    $self->log(5, "read_file: $total items read from $fn") if $self->{LogTo};

    return $conf;
}
   
# given hash ref, writes to file key = val
# NOTE: key should not contain '=';
# 
sub write_file {
    my ($self,$fn,$hash,$append,$header)  = @_;

    $fn = ">$fn" if $append;

    open CONF, ">$fn" || 
        return $self->error("Can't write file $fn: $!");
            
    print CONF "$header\n" if $header;
    foreach (sort keys %$hash) {
        return $self->error("Key cannot contain '=': $_") if /=/;
        printf CONF "%-22s = ", $_;
        if (ref($hash->{$_}) eq "ARRAY") {
            print CONF join(',', @{$hash->{$_}}) ."\n";
        } else {
            print CONF $hash->{$_} ."\n";
        }
    }
    close CONF;

    $self->log(5, "wrote to file: $fn");

    return 1;
}



sub default_server_conf {
    my $self = shift;
    my $defaults = {
        srl   => 1,
        bql   => 4,
        bqs   => 64,
        se    => '0F',  # engines 1, 2, 3, 4
        dre   => 4,
    };

    # split strings with , into array
    foreach (keys %$defaults) {
        $defaults->{$_} = $self->parse_value($defaults->{$_});
    }
    return $defaults;
}


sub default_agent_conf {
    my $self = shift;
    # 
    # These get overwritten by whatever's in config file,
    # which in turn gets overwritten by cmd-line options.
    # 
    my $defaults = {
        debuglevel         => "5",
        logfile            => "razor-agent.log",
        listfile_catalogue => "servers.catalogue.lst",
        listfile_nomination => "servers.nomination.lst",
        listfile_discovery => "servers.discovery.lst",
        min_cf             => "ac",
        turn_off_discovery => "0",
        ignorelist         => "0",
        razorzone          => "razor2.cloudmark.com",
        rediscovery_wait   => "172800",
        report_headers     => "1",
        whitelist          => "razor-whitelist",
        use_engines        => "1, 3, 4",
    };

    # 'razorhome' can exist in .conf, but we compute it instead of listing it here
    # 'rlimit' ?

    # split strings with , into array
    foreach (keys %$defaults) {
        $defaults->{$_} = $self->parse_value($defaults->{$_});
    }
    return $defaults;
}


1;

